(*
Author:  Alexander Lochmann <alexander.lochmann@uibk.ac.at> (2018)
License: LGPL (see file COPYING.LESSER)
*)
theory Ackbo
  imports AC_Weight
begin

locale weight_prec_ac =
  prec pr_strict +
  weight_fun w w0
  for w  :: "'f \<times> nat \<Rightarrow> nat"
    and w0 :: "nat"
    and pr_strict :: "('f \<times> nat) \<Rightarrow> ('f \<times> nat) \<Rightarrow> bool" +
  fixes AC ::"'f set"
begin

fun ackbo_case3 :: "(('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    "ackbo_case3 ord (Fun f ts) (Fun g ys) =
      (if f = g \<and> length ts = 2 \<and> length ys = 2 \<and> f \<in> AC then
        let (S, T) = (actop f (Fun f ts), actop f (Fun g ys)) in
        let s_rel = {(x, y). ord x y} in
        ackbo_case3_filtered_rel S T (f, length ts) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) s_rel
      else False)"
  | "ackbo_case3 _ _ _ = False"

lemma ackbo_case3_arg_intro:
  assumes "ackbo_case3 str s t"
  shows "\<exists> f s1 s2 t1 t2. s = Fun f [s1, s2] \<and> t = Fun f [t1, t2]"
  using ackbo_case3.elims(2) assms
  by (smt Bin_cases_with_length Term.term.simps(4) term.sel(4)) 

lemma ackbo_case3_local_mono:
  assumes "ackbo_case3 str (Fun f [s1, s2]) t"
  and "\<And> x y. x \<in># actop f (Fun f [s1, s2]) \<Longrightarrow> y \<in># actop f t \<Longrightarrow> str x y \<Longrightarrow> str' x y"
shows "ackbo_case3 str' (Fun f [s1, s2]) t"
proof -
 obtain t1 t2 where terms:"t = Fun f [t1, t2]" using assms(1) ackbo_case3_arg_intro by blast
 let ?S = "actop f (Fun f [s1,s2])"
 let ?T = "actop f (Fun f [t1, t2])"
 let ?f = "(f,2)" 
 let ?s_rel = "{(x, y). str x y}"
 let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" 
 consider (a) "filtered_nless_rel_s ?S ?T ?f ?ns_rel ?s_rel" |
          (b) "filtered_nless_rel_ns ?S ?T ?f ?ns_rel ?s_rel \<and> size ?S > size ?T"|
          (c) "filtered_nless_rel_ns ?S ?T ?f ?ns_rel ?s_rel \<and> size ?S = size ?T \<and>
           filtered_less_rel_s ?S ?T ?f ?ns_rel ?s_rel" by (smt ackbo_case3.simps(1) assms(1) case_prodD terms)
  then show ?thesis
  proof cases
    case a
    then have "filtered_nless_rel_s (actop f (Fun f [s1, s2])) (actop f (Fun f [t1, t2])) ?f ?ns_rel ?s_rel"
      using filtered_nless_rel_s_local_mono assms(2) by (smt case_prodD case_prodI mem_Collect_eq terms) 
    then have "ackbo_case3 str' (Fun f [s1,s2]) t" using terms
      by (smt ackbo_case3.simps(1) assms(1) assms(2) case_prod_conv filtered_nless_rel_s_local_mono mem_Collect_eq) 
    then show ?thesis by blast
  next
    case b
    then have "ackbo_case3 str' (Fun f [s1,s2]) t" using filtered_nless_rel_ns_local_mono terms
      by (smt ackbo_case3.simps(1) assms(1) assms(2) case_prodD case_prodI mem_Collect_eq) 
    then show ?thesis by blast
  next
    case c
    then have "ackbo_case3 str' (Fun f [s1,s2]) t" using filtered_nless_rel_ns_local_mono filtered_less_rel_s_local_mono terms assms(2)
      by (smt ackbo_case3.simps(1) assms(1) case_prodD case_prodI mem_Collect_eq)
    then show ?thesis by blast
  qed
qed

lemma ackbo_case3_pres_root:
  fixes x :: "('f, 'v) term"
  assumes "ackbo_case3 ord x y"
  shows "root x = root y"
  using assms by (smt ackbo_case3.elims(2) root.elims term.distinct(1) term.inject(2))

lemma ackbo_case3_num_args:
  assumes "ackbo_case3 ord x y"
  shows "num_args x = 2 \<and> num_args y = 2"
proof -
  obtain s ss where wit_term_x:"x = Fun s ss" using assms
    using ackbo_case3.elims(2) by blast 
  then have "length ss = 2" using assms by (metis ackbo_case3.simps(1) ackbo_case3.simps(3) root.elims) 
  obtain t ts where wit_term_y:"y = Fun t ts" using assms
    using ackbo_case3.elims(2) by blast 
  then have "length ts = 2" using assms by (metis ackbo_case3.simps(1) wit_term_x) 
  then show ?thesis using assms wit_term_x wit_term_y by (simp add: \<open>length ss = 2\<close>) 
qed

lemma ackbo_case3_num_args_2:
  fixes x ::"('f, 'v)term"
  assumes "num_args x \<noteq> 2 \<or> num_args y \<noteq> 2"
  shows "\<not> ackbo_case3 ord x y"
  using assms
proof
  assume "num_args x \<noteq> 2"
  then show ?thesis using ackbo_case3_num_args by blast 
next
  assume "num_args y \<noteq> 2"
  then show ?thesis using ackbo_case3_num_args by blast
qed

lemma ackbo_case3_mono [mono]:
  assumes "\<And>x y. ord x y \<longrightarrow> ord' x y"
  shows "ackbo_case3 ord s t \<longrightarrow> ackbo_case3 ord' s t"
proof
  assume "ackbo_case3 ord s t"
  then show "ackbo_case3 ord' s t"
  using assms
      and ns_mul_ext_mono[where ns="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and ns'="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and s="{(x, y). ord x y}" and s'="{(x, y). ord' x y}"]
      and s_mul_ext_mono[where ns="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and ns'="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and s="{(x, y). ord x y}" and s'="{(x, y). ord' x y}"]
  apply (cases "(ord, s, t)" rule:ackbo_case3.cases) apply (auto split: if_splits)
  by (smt Collect_mono_iff case_prodE case_prod_conv subsetCE)
qed

lemma ackbo_case3_cong [fundef_cong]:
  assumes "s = u"
    and "t = v"
    and "\<And> x y. u \<rhd> x \<Longrightarrow> v \<rhd> y \<Longrightarrow> R x y = S x y"
  shows "ackbo_case3 R s t = ackbo_case3 S u v"
  using assms by (smt ackbo_case3_arg_intro ackbo_case3_local_mono actop_mset_elem_subterm) 

function (sequential) ackbo_impl :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    "ackbo_impl s t =
      (if vars_term_ms t \<subseteq># vars_term_ms s \<and> weight t \<le> weight s then
        if weight t < weight s then True else
        (case s of
          Var y \<Rightarrow> False 
        | Fun f ss \<Rightarrow>
          (case t of
            Var x \<Rightarrow> True
          | Fun g ts \<Rightarrow>
            if pr_strict (f,length ss) (g,length ts) then True
            else if (f, length ss) = (g, length ts) then
              if f \<notin> AC \<or> length ss \<noteq> 2 then
                fst (lex_ext (\<lambda>x y. (ackbo_impl x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length ss) ss ts)
              else ackbo_case3 ackbo_impl s t
            else False))
      else False)"
  by pat_completeness auto
termination
  apply (relation "measure (\<lambda> (s,t). term_size s)")
    apply (auto dest: supt_term_size intro!: le_imp_less_Suc)
   apply (metis elem_le_sum_list length_map nth_map)+
  done

end
sublocale weight_prec_ac \<subseteq> weight_fun .
sublocale admissible_weight_fun \<subseteq> weight_prec_ac .

locale admissible_weight_fun_ac =
  admissible_weight_fun w w0 pr_strict 
  for w  :: "'f \<times> nat \<Rightarrow> nat"
    and w0 :: "nat"
    and pr_strict :: "('f \<times> nat) \<Rightarrow> ('f \<times> nat) \<Rightarrow> bool" +
  fixes AC ::"'f set"
begin

inductive ackbo_case0 :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    "ackbo_case0 (Fun f [Var x]) (Var x)" |
    "ackbo_case0 (Fun f ts) x \<Longrightarrow> ackbo_case0 (Fun f [Fun f ts]) x"

lemma ackbo_case0_subterm:
  "ackbo_case0 s t \<Longrightarrow> s \<rhd> t"
  by (induct rule:ackbo_case0.induct) auto

lemma ackbo_case0_rhs_var:
  assumes "ackbo_case0 s t"
  shows "is_Var t"
  using assms
  by (induct rule:ackbo_case0.induct) auto

lemma ackbo_case0_lhs_length:
  assumes "ackbo_case0 (Fun s ss) t"
  shows "length ss = 1"
  using assms by (metis One_nat_def Suc_eq_plus1 ackbo_case0.cases list.size(3) list.size(4) term.sel(4)) 

fun ackbo_case1 :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    "ackbo_case1 (Fun f ts) (Fun g ys) = pr_strict (f, length ts) (g, length ys)"|
    "ackbo_case1 other wise = False"

lemma ackbo_case1_lhs_Fun:
  assumes "ackbo_case1 s t"
  shows "is_Fun s"
  using assms by auto 

lemma ackbo_case1_rhs_Fun:
  assumes "ackbo_case1 s t"
  shows "is_Fun t"
  using assms by auto 

lemma ackbo_case1_different_Fun:
  assumes "ackbo_case1 (Fun f ts) (Fun f ss)"
  and "length ts = length ss"
  shows "f \<noteq> s"
proof (rule ccontr)
  assume "\<not> f \<noteq> s"
  then have a:"f = s" by blast
  from assms have "pr_strict (f ,length ts) (s , length ss)" by (simp add: a)
  then have "pr_strict (f ,length ts) (f , length ss)" by (simp add:a)
  then show False using pr_irr assms by auto
qed

lemma ackbo_case0_lhs_contains_var:
  assumes "ackbo_case0 s t"
  shows "vars_term_ms t \<subseteq># vars_term_ms s"
  using assms
  by (induct rule:ackbo_case0.induct) auto

lemma ackbo_case0_rhs_contains_var:
  assumes "ackbo_case0 s t"
  shows "vars_term_ms s \<subseteq># vars_term_ms t"
  using assms by (induct rule: ackbo_case0.induct) auto

fun ackbo_case2 :: "(('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    "ackbo_case2 ord (Fun f ts) (Fun g ys) =
      (if f = g \<and> length ts = length ys \<and> (length ts \<noteq> 2 \<or> f \<notin> AC) then
        fst (lex_ext (\<lambda> x y. (ord x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length ts) ts ys)
      else False)"
  | "ackbo_case2 _ other wise = False"

lemma ackbo_case2_single_arg:
  assumes "ackbo_case2 ord (Fun f [f1]) (Fun f [g1])"
  shows "ord f1 g1"
  using assms by (simp add: lex_ext_iff)

lemma ackbo_case2_pres_root:
  fixes x :: "('f, 'v) term"
  assumes "ackbo_case2 ord x y"
  shows "root x = root y"
  using assms by (smt ackbo_case2.elims(2) admissible_weight_fun_axioms root.elims term.distinct(1) term.inject(2))

lemma ackbo_case2_mono [mono]:
  assumes "\<And>s t. ord s t \<longrightarrow> ord' s t"
  shows "ackbo_case2 ord s t \<longrightarrow> ackbo_case2 ord' s t"
proof
  assume "ackbo_case2 ord s t"
  then show "ackbo_case2 ord' s t"
    using assms and lex_ext_to_lex_ext_unbounded and lex_ext_unbounded_mono[where P="(\<lambda> x y. (ord x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*))" and P'="(\<lambda> x y. (ord' x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*))"]
    apply (cases "(ord, s, t)" rule: ackbo_case2.cases)
    apply (auto split: if_splits)
    apply (simp add: lex_ext_def)
    by (simp add: lex_ext_def)
qed

lemma ackbo_case2_local_mono [mono]:
  assumes "ackbo_case2 ord (Fun f ss) (Fun f ts)"
  assumes "\<And>s t. s \<in> set ss \<Longrightarrow> t \<in> set ts \<Longrightarrow> ord s t \<Longrightarrow> ord' s t"
  shows "ackbo_case2 ord' (Fun f ss) (Fun f ts)" 
    using assms and lex_ext_to_lex_ext_unbounded and lex_ext_unbounded_mono[where P="(\<lambda> x y. (ord x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*))" and P'="(\<lambda> x y. (ord' x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*))"]
    apply (cases "(ord, Fun f ss, Fun f ts)" rule: ackbo_case2.cases)
    apply (auto split: if_splits)
    apply (simp add: lex_ext_def)
    by (simp add: lex_ext_def)

lemma ackbo_case3_mono [mono]:
  assumes "\<And>x y. ord x y \<longrightarrow> ord' x y"
  shows "ackbo_case3 ord s t \<longrightarrow> ackbo_case3 ord' s t"
proof
  assume "ackbo_case3 ord s t"
  then show "ackbo_case3 ord' s t"
  using assms
      and ns_mul_ext_mono[where ns="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and ns'="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and s="{(x, y). ord x y}" and s'="{(x, y). ord' x y}"]
      and s_mul_ext_mono[where ns="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and ns'="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" and s="{(x, y). ord x y}" and s'="{(x, y). ord' x y}"]
  apply (cases "(ord, s, t)" rule:ackbo_case3.cases) apply (auto split: if_splits)
  by (smt Collect_mono_iff case_prodE case_prod_conv subsetCE)
qed

inductive ackbo :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    case_w: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s > weight t \<Longrightarrow> ackbo s t" |
    case_0: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow> ackbo_case0 s t \<Longrightarrow> ackbo s t" |
    case_1: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow> ackbo_case1 s t \<Longrightarrow> ackbo s t" |
    case_2: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow> ackbo_case2 ackbo s t \<Longrightarrow> ackbo s t" |
    case_3: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow> ackbo_case3 ackbo s t \<Longrightarrow> ackbo s t"

lemma ackbo_rhs_var_imp_case0:
  assumes "ackbo s t"
  and "weight s = weight t"
  and "is_Var t"
  shows "ackbo_case0 s t"
proof -
  have "\<not> ackbo_case1 s t" using assms(3) ackbo_case1_rhs_Fun by blast
  moreover have "\<not> ackbo_case2 ackbo s t" using assms(3) by auto
  moreover have "\<not> ackbo_case3 ackbo s t" using assms(3) by auto
  ultimately show ?thesis using ackbo.cases assms  by fastforce 
qed

lemma ackbo_weight_ineq:
  assumes "ackbo s t"
    and "weight s \<noteq> weight t"
  shows "weight s > weight t"
proof (rule ccontr)
  assume "\<not> weight t < weight s"
  then have "weight s < weight t" using assms(2) by arith
  then have "\<not> ackbo s t" using ackbo.simps[of s t] by auto
  thus "False" using assms(1) by simp
qed

lemma ackbo_weight_eq:
  assumes "ackbo s t"
    and "\<not> weight s > weight t"
  shows "weight s = weight t"
proof (rule ccontr)
  assume "weight s \<noteq> weight t"
  then have "weight s > weight t" using assms(1) using ackbo_weight_ineq by blast
  thus "False" using assms(2) by simp
qed

lemma ackbo_weight_pos:
  assumes "ackbo s t"
  shows "weight s = weight t \<or> weight s > weight t"
  using assms by (induct rule: ackbo.induct) auto

lemma ackbo_case_0_lhs_subterm_exists:
  assumes "ackbo_case0 (Fun f ts) t"
  shows "\<exists> u. ts = [u]"
  using assms ackbo_case0.cases by blast 

lemma ackbo_case_0_subterm_lhs_var_or_same_root:
  assumes "ackbo_case0 (Fun f [u]) t"
  shows "is_Var u \<or> root (Fun f [u]) = root u"
  using assms
proof (induct "(Fun f [u])" t rule: ackbo_case0.induct)
  case (1 x)
  then show ?case by simp
next
  case (2 ts x)
  then show ?case using ackbo_case_0_lhs_subterm_exists by fastforce 
qed

lemma ackbo_case2_impl_lex_ext_s:
  assumes "ackbo_case2 ord (Fun f ss) (Fun f ts)"
  shows "fst (lex_ext (\<lambda> x y. (ord x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length ss) ss ts)"
  by (smt ackbo_case2.elims(2) admissible_weight_fun_axioms assms lex_ext_cong term.sel(4))

lemma ackbo_case2_arg_intro:
  assumes "ackbo_case2 f s t"
  shows "\<exists> f ss ts. s = Fun f ss \<and> t = Fun f ts"
  using ackbo_case2.elims(2) assms by fastforce

lemma ackbo_case2_length_eq:
  assumes "ackbo_case2 f (Fun g ss) (Fun g ts)"
  shows "length ss = length ts"
  by (meson ackbo_case2.simps(1) assms)

lemma ackbo_case2_impl_not_acsym:
  assumes "ackbo_case2 f (Fun g ss) (Fun g ts)"
  shows "g \<notin> AC \<or> length ss \<noteq> 2"
  by (meson ackbo_case2.simps(1) assms)

lemma ackbo_case2_to_witness:
  assumes "ackbo_case2 f (Fun g ss) (Fun g ts)"
  shows "\<exists> i < length ss. i < length ts \<and> (\<forall> j < i. snd ((\<lambda> x y.(f x y , (x,y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (ss ! j) (ts ! j)))
                                           \<and> fst ((\<lambda> x y. (f x y, (x,y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (ss ! i) (ts ! i))"
  using assms ackbo_case2_length_eq[of f g ss ts] ackbo_case2.simps(1)[of f g ss g ts] ackbo_case2_impl_not_acsym[of f g ss ts]
    lex_ext_iff[of "\<lambda> x y. (f x y, (x,y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" "length ss" ss ts]
    by (smt eq_fst_iff less_not_refl3)

lemma lex_ext_to_ackbo_case2:
  assumes "\<exists> i < length ss. i < length ts \<and> (\<forall> j < i. snd ((\<lambda> x y.(f x y , (x,y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (ss ! j) (ts ! j)))
                              \<and> fst ((\<lambda> x y. (f x y, (x,y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (ss ! i) (ts ! i))"
  and "s = Fun g ss"
  and "t = Fun g ts"
  and "g \<notin> AC"
  and "length ss = length ts"
shows "ackbo_case2 f s t"
  using assms by (simp add: lex_ext_iff)

lemma order_pair_impl_ackbo_case2_trans:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  and "ackbo_case2 str s t"
  and "ackbo_case2 str t u"
shows "ackbo_case2 str s u"
proof -
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  obtain f ss ts us where terms:"s = Fun f ss \<and> t = Fun f ts \<and> u = Fun f us" using ackbo_case2_arg_intro assms(2-3)
    by (metis term.inject(2))
  then have tr:"trans {(x, y). fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) x y)}"
    using order_pair.lex_ext_order_pair assms order_pair.axioms pre_order_pair_def by fastforce
  moreover have "length ss = length ts" using ackbo_case2_length_eq[of str f ss ts] assms(2) terms by blast 
  ultimately have "fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) ss us)"
    unfolding trans_def using assms(2-3) ackbo_case2_impl_lex_ext_s terms by fastforce
  then show ?thesis using terms by (smt \<open>length ss = length ts\<close> ackbo_case2.simps(1) assms(3) lex_ext_cong)
qed

lemma order_pair_impl_ackbo_case2_l_comp:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "(s, t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_case2 str t u"
  shows "ackbo_case2 str s u"
proof -
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  obtain f ts us where terms_tu:"t = Fun f ts \<and> u = Fun f us" using ackbo_case2_arg_intro assms(3) by blast
  then obtain ss where term_s:"s = Fun f ss" using assms(2) by (metis ac_terms_eq_root fst_conv root.simps(2) root_Some)
  then have comp:"compat_pair {(x, y). fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) x y)}
                              {(x, y). snd (lex_ext (\<lambda> l r. (str l r , (l, r) \<in> ?ns_rel)) (length ss) x y)}"
    using order_pair.lex_ext_order_pair assms order_pair.axioms by fastforce
  have eq:"\<And> x y .str x y \<longleftrightarrow> (x,y) \<in> {(x,y). str x y}" by blast
  then have l:"(f \<notin> AC \<or> length ss \<noteq> 2)\<and> length ss = length ts" using ac_terms_eq_root assms(2-3) term_s terms_tu by fastforce
  then have "snd (lex_ext (\<lambda> l r. (str l r , (l, r) \<in> ?ns_rel)) (length ss) ss ts)"
    using term_s terms_tu assms(2) eq acstep_no_ac_root_impl_acstep_lex_ext[of s t AC f ss ts] by (smt lex_ext_cong)
  moreover have "fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) ts us)"
    using assms(3) ackbo_case2_impl_lex_ext_s[of str f ts us] terms_tu l by auto
  ultimately have "fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) ss us)"
    using comp unfolding compat_pair_def using terms_tu term_s by fastforce
  then show ?thesis using term_s terms_tu l by (smt ackbo_case2.simps(1) assms(3) lex_ext_cong) 
qed

lemma order_pair_impl_ackbo_case2_r_comp:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_case2 str s t"
    and "(t, u) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "ackbo_case2 str s u"
proof -
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  obtain f ss ts where terms_st:"s = Fun f ss \<and> t = Fun f ts" using ackbo_case2_arg_intro assms(2) by blast
  then obtain us where term_u:"u = Fun f us" using assms(3) by (metis ac_terms_eq_root fst_conv root.simps(2) root_Some)
  then have comp:"compat_pair {(x, y). fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) x y)}
                              {(x, y). snd (lex_ext (\<lambda> l r. (str l r , (l, r) \<in> ?ns_rel)) (length ss) x y)}"
    using order_pair.lex_ext_order_pair assms order_pair.axioms by fastforce
  have eq:"\<And> x y .str x y \<longleftrightarrow> (x,y) \<in> {(x,y). str x y}" by blast
  have l:"(f \<notin> AC \<or> length ss \<noteq> 2) \<and> length ss = length ts" using ackbo_case2.simps(1) assms(2) terms_st by metis
  have "length ts = length us" using ac_terms_eq_root assms(3) terms_st term_u by fastforce
  then have "snd (lex_ext (\<lambda> l r. (str l r , (l, r) \<in> ?ns_rel)) (length ss) ts us)" 
    using terms_st term_u assms(3) eq acstep_no_ac_root_impl_acstep_lex_ext[of t u AC f ts us "{(x, y). str x y}"] l by auto
  moreover have "fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) ss ts)"
    using assms(2) ackbo_case2_impl_lex_ext_s[of str f ss ts] terms_st l by auto
  ultimately have "fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) (length ss) ss us)"
    using comp unfolding compat_pair_def using terms_st term_u by fastforce
  then show ?thesis using terms_st term_u l \<open>length ts = length us\<close> by auto  
qed

lemma order_pair_impl_ackbo_case2_order_pair:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "order_pair {(x,y). ackbo_case2 str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
proof
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  have "str l r \<longleftrightarrow> (l,r) \<in> {(x,y). str x y}" by blast
  then have ord:"order_pair {(x, y). fst (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) n x y)}
                       {(x, y). snd (lex_ext (\<lambda> l r. (str l r, (l, r) \<in> ?ns_rel)) n x y)}"
    using order_pair.lex_ext_order_pair assms by fastforce
  show "refl ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" by (simp add: conversion_def refl_rtrancl) 
  show "trans ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" by (simp add: conversion_trans)
  show "trans {(x, y). ackbo_case2 str x y}" unfolding trans_def using order_pair_impl_ackbo_case2_trans assms(1) by blast
  show "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* O {(x, y). ackbo_case2 str x y} \<subseteq> {(x, y). ackbo_case2 str x y}" using order_pair_impl_ackbo_case2_l_comp assms(1) by blast
  show "{(x, y). ackbo_case2 str x y} O (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> {(x, y). ackbo_case2 str x y}" using order_pair_impl_ackbo_case2_r_comp assms(1) by blast
qed


lemma ackbo_case3_Elim:
  assumes "ackbo_case3 ord (Fun f ss) t"
  shows "ackbo_case3_filtered_rel (actop f (Fun f ss))  (actop f t) (f,2) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) {(x, y). ord x y}"
  using assms by (smt ackbo_case3.elims(2) case_prodD term.inject(2)) 

lemma order_pair_impl_ackbo_case3_trans:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_case3 str s t"
    and "ackbo_case3 str t u"
  shows "ackbo_case3 str s u"
proof -
  let ?str_rel = "{(x,y). str x y}"
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  obtain f s1 s2 t1 t2 u1 u2 where terms:"s = Fun f [s1,s2] \<and> t = Fun f [t1,t2] \<and> u = Fun f [u1,u2]" using assms(2-3)
    by (smt ackbo_case3_arg_intro term.inject(2))
  let ?S = "actop f s"
  let ?T = "actop f t"
  let ?U = "actop f u"
  have ord:"order_pair {(S,T). filtered_nless_rel_s S T (f,2) ?ns_rel ?str_rel} {(S,T). filtered_nless_rel_ns  S T (f,2) ?ns_rel ?str_rel}"
      using A2 assms(1) by blast
  have t_cases:"ackbo_case3_filtered_rel ?T ?U (f,2) ?ns_rel ?str_rel"
    using assms(3) ackbo_case3_Elim[of str f] terms by fastforce
  consider (a) "filtered_nless_rel_s ?S ?T (f,2) ?ns_rel ?str_rel" |
           (b) "filtered_nless_rel_ns ?S ?T (f,2) ?ns_rel ?str_rel \<and> size ?S > size ?T" |
           (c) "filtered_nless_rel_ns ?S ?T (f,2) ?ns_rel ?str_rel \<and> size ?S = size ?T \<and> 
                filtered_less_rel_s ?S ?T (f,2) ?ns_rel ?str_rel"
    using assms(2) ackbo_case3_Elim[of str f] terms by fastforce
  then show ?thesis
  proof cases
    case a
    from t_cases have "filtered_nless_rel_ns ?T ?U (f,2) ?ns_rel ?str_rel  \<or> filtered_nless_rel_s ?T ?U (f,2) ?ns_rel ?str_rel " by auto
    then have "filtered_nless_rel_s ?S ?U (f,2) ?ns_rel ?str_rel"
      using a by (smt assms(1) filter_compact_r filter_trans_s)
    then show ?thesis using assms(2-3) terms by (smt ackbo_case3.simps(1) case_prodI)
  next
    case b
    from t_cases have "filtered_nless_rel_ns ?T ?U (f,2) ?ns_rel ?str_rel  \<and> size ?S > size ?U \<or> filtered_nless_rel_s ?T ?U (f,2) ?ns_rel ?str_rel "
      using b by auto
    then have "filtered_nless_rel_s ?S ?U (f,2) ?ns_rel ?str_rel \<or> filtered_nless_rel_ns ?S ?U (f,2) ?ns_rel ?str_rel \<and> size ?S > size ?U"
      using b filter_compact_l filter_trans_ns assms(1) by blast
    then show ?thesis using assms(2-3) terms ackbo_case3.simps(1) by (smt case_prodI)
  next
    case c
    then have "filtered_nless_rel_s ?S ?U (f,2) ?ns_rel ?str_rel \<or> filtered_nless_rel_ns ?S ?U (f,2) ?ns_rel ?str_rel \<and> size ?S > size ?U \<or>
               filtered_nless_rel_ns ?S ?U (f,2) ?ns_rel ?str_rel \<and> size ?S = size ?U \<and> filtered_less_rel_s ?S ?U (f,2) ?ns_rel ?str_rel"
      using c t_cases filter_compact_l filter_trans_ns filtered_less_rel_s_cond_trans assms(1)
      by (metis (no_types, lifting)) 
  then show ?thesis using assms(2-3) terms ackbo_case3.simps(1) by (smt case_prodI)
  qed
qed

lemma order_pair_impl_ackbo_case3_l_comp:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "(s, t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_case3 str t u"
  shows "ackbo_case3 str s u"
proof -
  let ?str_rel = "{(x,y). str x y}"
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  obtain f t1 t2 u1 u2 where terms_tu:"t = Fun f [t1, t2] \<and> u = Fun f [u1, u2]" using ackbo_case3_arg_intro assms(3) by blast
  then have ac_sym:"f \<in> AC" using assms(3) by (smt ackbo_case3.simps(1))
  let ?S = "actop f s"
  let ?T = "actop f t"
  let ?U = "actop f u"
  have "root s = root t" using assms(2) ac_terms_eq_root[of s t AC] by blast
  then obtain ss where "s = Fun f ss \<and> length ss = 2" using terms_tu root.simps(2)[of f "[t1,t2]"] root_Some[of s]
    by (metis One_nat_def fst_conv length_Cons list.size(3) list.size(4) nat_1_add_1 snd_conv) 
  then obtain s1 s2 where term_s:"s = Fun f [s1, s2]" by (metis Bin_cases_with_length term.distinct(1) term.inject(2))
  have eq_size:"size ?S = size ?T"
    using multpw_impl_eq_size acstep_impl_actop_multpw assms(2) terms_tu term_s ac_sym by (smt list.size(4))
  have ns:"filtered_nless_rel_ns ?S ?T (f,2) ?ns_rel ?str_rel" using assms(2) acstep_impl_filtered_nless_rel_ns[of s t] terms_tu term_s ac_sym
    by (metis \<open>s = Fun f ss \<and> length ss = 2\<close> ackbo_case3_num_args_2 assms(3) term.sel(4))
  consider (a) "filtered_nless_rel_s ?T ?U (f,2) ?ns_rel ?str_rel" |
           (b) "filtered_nless_rel_ns ?T ?U (f,2) ?ns_rel ?str_rel  \<and> size ?T > size ?U" |
           (c) "filtered_nless_rel_ns ?T ?U (f,2) ?ns_rel ?str_rel  \<and> size ?T = size ?U \<and> 
                filtered_less_rel_s ?T ?U (f,2) ?ns_rel ?str_rel"
    using assms(3) ackbo_case3_Elim[of str f] terms_tu by fastforce
  then show ?thesis
  proof cases
    case a
    then have "filtered_nless_rel_s ?S ?U (f, 2) ?ns_rel ?str_rel"
      using filter_compact_l assms(1) ns by auto
    then show ?thesis using terms_tu term_s list.size(4) by (smt ackbo_case3.simps(1) assms(3) case_prodI) 
  next
    case b
    then have "filtered_nless_rel_ns ?S ?U (f, 2) ?ns_rel ?str_rel \<and> size ?U < size ?S"
      using filter_trans_ns eq_size ns assms(1) by (metis (no_types, lifting))
    then show ?thesis using terms_tu term_s list.size(4) by (smt ackbo_case3.simps(1) assms(3) case_prodI)
  next
    case c
    then have "filtered_nless_rel_ns ?S ?U (f, 2) ?ns_rel ?str_rel \<and> size ?U = size ?S \<and>
               filtered_less_rel_s ?S ?U (f, 2) ?ns_rel ?str_rel"
      using filter_trans_ns eq_size ns assms(1) acstep_filter_r_compact_r
      by (metis (no_types, lifting) \<open>s = Fun f ss \<and> length ss = 2\<close> ac_sym assms(2)) 
    then show ?thesis using terms_tu term_s list.size(4) by (smt ackbo_case3.simps(1) assms(3) case_prodI)
  qed
qed

lemma order_pair_impl_ackbo_case3_r_comp:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_case3 str s t"
    and "(t, u) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "ackbo_case3 str s u"
proof -
  let ?str_rel = "{(x,y). str x y}"
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  obtain f s1 s2 t1 t2 where terms_st:"s = Fun f [s1, s2] \<and> t = Fun f [t1, t2]" using ackbo_case3_arg_intro assms(2) by blast
  then have ac_sym:"f \<in> AC" using assms(2) by (smt ackbo_case3.simps(1))
  let ?S = "actop f s"
  let ?T = "actop f t"
  let ?U = "actop f u"
  have "root t = root u" using assms(3) ac_terms_eq_root[of t u AC] by blast
  then obtain us where "u = Fun f us \<and> length us = 2" using terms_st root.simps(2)[of f "[t1,t2]"] root_Some[of u]
    by (metis One_nat_def fst_conv length_Cons list.size(3) list.size(4) nat_1_add_1 snd_conv) 
  then obtain u1 u2 where term_u:"u = Fun f [u1, u2]" by (metis Bin_cases_with_length term.distinct(1) term.inject(2))
  have eq_size:"size ?T = size ?U"
    using multpw_impl_eq_size acstep_impl_actop_multpw assms(3) terms_st term_u ac_sym by (smt list.size(4))
  have ns:"filtered_nless_rel_ns ?T ?U (f,2) ?ns_rel ?str_rel " using assms(2) acstep_impl_filtered_nless_rel_ns terms_st term_u ac_sym
    by (smt \<open>u = Fun f us \<and> length us = 2\<close> assms(3) length_Cons term.inject(2))
  consider (a) "filtered_nless_rel_s ?S ?T (f,2) ?ns_rel ?str_rel" |
           (b) "filtered_nless_rel_ns ?S ?T (f,2) ?ns_rel ?str_rel \<and> size ?S > size ?T" |
           (c) "filtered_nless_rel_ns ?S ?T (f,2) ?ns_rel ?str_rel \<and> size ?S = size ?T \<and> 
                filtered_less_rel_s ?S ?T (f,2) ?ns_rel ?str_rel"
    using assms(2) ackbo_case3_Elim[of str f] terms_st by fastforce
  then show ?thesis
  proof cases
    case a
    then have "filtered_nless_rel_s ?S ?U (f, 2) ?ns_rel ?str_rel"
      using filter_compact_r assms(1) ns by auto
    then show ?thesis using terms_st term_u list.size(4) by (smt ackbo_case3.simps(1) assms(2) case_prodI) 
  next
    case b
    then have "filtered_nless_rel_ns ?S ?U (f, 2) ?ns_rel ?str_rel \<and> size ?U < size ?S"
      using filter_trans_ns eq_size ns assms(1) by (metis (no_types, lifting))
    then show ?thesis using terms_st term_u list.size(4) by (smt ackbo_case3.simps(1) assms(2) case_prodI)
  next
    case c
    then have "filtered_nless_rel_ns ?S ?U (f, 2) ?ns_rel ?str_rel \<and> size ?U = size ?S \<and>
               filtered_less_rel_s ?S ?U (f, 2) ?ns_rel ?str_rel"
      using filter_trans_ns eq_size ns assms(1) acstep_filter_r_compact_l
      by (smt \<open>u = Fun f us \<and> length us = 2\<close> ac_sym assms(3) list.size(4) term.inject(2) term_u terms_st)  
    then show ?thesis using terms_st term_u list.size(4) by (smt ackbo_case3.simps(1) assms(2) case_prodI)
  qed
qed

lemma order_pair_impl_ackbo_case3_order_pair:
  fixes str:: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" 
  assumes "order_pair {(x,y). str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "order_pair {(x,y). ackbo_case3 str x y} ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
proof
  let ?ns_rel = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f,'v) term rel"
  have "str l r \<longleftrightarrow> (l,r) \<in> {(x,y). str x y}" by blast
  show "refl ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" by (simp add: conversion_def refl_rtrancl) 
  show "trans ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" by (simp add: conversion_trans)
  show "trans {(x, y). ackbo_case3 str x y}" unfolding trans_def using order_pair_impl_ackbo_case3_trans assms(1) by blast
  show "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* O {(x, y). ackbo_case3 str x y} \<subseteq> {(x, y). ackbo_case3 str x y}" using order_pair_impl_ackbo_case3_l_comp assms(1) by blast
  show "{(x, y). ackbo_case3 str x y} O (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> {(x, y). ackbo_case3 str x y}" using order_pair_impl_ackbo_case3_r_comp assms(1) by blast
qed


lemma ackbo_trivial_context_closure:
  assumes "ackbo s t"
  shows "ackbo (Fun f [s]) (Fun f [t])"
proof -
  have vars:"vars_term_ms (Fun f [t]) \<subseteq># vars_term_ms (Fun f [s])" using assms ackbo.simps by auto
  consider (a) "weight s > weight t" | (b) "weight s = weight t" using assms ackbo.simps by blast
  then show ?thesis
  proof cases
    case a
    then have "weight (Fun f [s]) > weight (Fun f [t])" by auto
      then show ?thesis using vars ackbo.case_w by blast
  next
    case b
    then have "weight (Fun f [s]) = weight (Fun f [t])" by auto
    moreover have "fst (lex_ext (\<lambda> x y. (ackbo x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length [s]) [s] [t])" using assms by (simp add: lex_ext_iff) 
    ultimately show ?thesis using ackbo.case_2 ackbo_case2.simps(1)[of ackbo f "[s]" f "[t]"] vars by fastforce
  qed
qed

lemma ackbo_no_var_impl_prstrict_or_eq_root:
  assumes "ackbo s t"
    and "s = Fun f fs"
    and "t = Fun g gs"
    and "weight s = weight t"
  shows "pr_strict (f,length fs) (g,length gs) \<or> f = g \<and> length fs = length gs"
using assms
proof (cases)
  case case_w
  then show ?thesis using assms(4) by auto
next
  case case_0
  then have "is_Var t" using ackbo_case0_rhs_var by blast
  then show ?thesis using assms(3) by auto
next
  case case_1
  then show ?thesis using ackbo_case1.simps(1) assms by blast
next
  case case_2
  then show ?thesis using ackbo_case2.simps(1) assms by force
next
  case case_3
  then show ?thesis using ackbo_case3.simps(1) assms by metis 
qed

lemma ackbo_acsym_imp_case3:
  assumes "ackbo s t"
    and "weight s = weight t"
    and "s = Fun f fs"
    and "t = Fun f gs"
    and "length fs = 2 \<and> length gs = 2"
    and "f \<in> AC"
  shows "ackbo_case3 ackbo s t"
proof -
  have "\<not> ackbo_case0 s t" using ackbo_case0_rhs_var assms(4) by blast
  moreover have "\<not> ackbo_case1 s t" using pr_irr ackbo_case1.simps(1)[of f fs f gs] assms(3-5) by fastforce
  moreover have "\<not> ackbo_case2 ackbo s t" using ackbo_case2.simps(1) assms(3-6) by fastforce
  ultimately show ?thesis using assms(1-2) by (simp add: ackbo.simps)
qed

lemma ackbo_eq_impl_left:
  assumes "ackbo s t"
  shows "ackbo_impl s t"
  using assms
proof induction
  case (case_w t s)
  then show ?case by simp
next
  case (case_0 t s)
  obtain x where t: "t = Var x" using ackbo_case0_rhs_var[OF case_0(3)] by blast
  from ackbo_case0.simps[of s] obtain f ss where "s = Fun f ss" using case_0(3) by auto 
  then show ?case using case_0 t by simp
next
  case (case_1 t s)
  obtain f g ss ts where s: "s = Fun f ss" and t: "t = Fun g ts" using case_1(3) by (metis ackbo_case1.elims(2)) 
  then have "pr_strict (f, length ss) (g, length ts)" using case_1(3) by (metis ackbo_case1.simps(1))
  then show ?case using s t case_1 by simp
next
  case (case_2 t s)
  obtain f ss ts where s: "s = Fun f ss" and t: "t = Fun f ts" using case_2(3) ackbo_case2_arg_intro by blast
  have "ackbo_case2 (ackbo_impl) s t" using case_2(3) by (metis (no_types, lifting) ackbo_case2_mono)
  then have "fst (lex_ext (\<lambda> x y. (ackbo_impl x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length ss) ss ts)"
    using s t ackbo_case2_to_witness[of "ackbo_impl" f ss ts] ackbo_case2_impl_lex_ext_s by blast
  moreover have "f \<notin> AC \<or> length ss \<noteq> 2" using case_2(3) ackbo_case2_impl_not_acsym s t by blast 
  moreover have "length ss = length ts" using case_2(3) ackbo_case2_length_eq s t by blast
  ultimately show ?case using case_2(1,2,3) s t by simp
next
  case (case_3 t s)
  obtain f ss ts where s: "s = Fun f ss" and t: "t = Fun f ts" using case_3(3) ackbo_case3_arg_intro by blast
  moreover have "f \<in> AC \<and> length ss = 2" using case_3(3) by (metis (no_types, lifting) ackbo_case3.simps(1) s t) 
  moreover have "length ss = length ts" by (metis (no_types, lifting) ackbo_case3.simps(1) case_3.IH s t)
  moreover have "ackbo_case3 (\<lambda>x1 x2. ackbo_impl x1 x2) s t" using case_3 by (metis (no_types, lifting) ackbo_case3_mono) 
  then show ?case using case_3 s t by (smt Term.term.simps(6) ackbo_case3.simps(1) ackbo_impl.elims(3) order_refl)
qed

lemma ackbo_eq_impl_right:
  assumes "ackbo_impl s t"
  shows "ackbo s t"
  using assms
proof (induct s arbitrary: t rule: subterm_induct)
  case (subterm u)
  have vars: "vars_term_ms t \<subseteq># vars_term_ms u" using subterm(2) by (meson weight_prec_ac.ackbo_impl.simps)
  have wei: "weight t \<le> weight u" using subterm(2) by (meson weight_prec_ac.ackbo_impl.simps)
  {
    assume as: "is_Var u"
    then have "weight u = w0" by fastforce
    moreover have "weight t \<ge> w0" by (simp add: weight_w0)
    ultimately have False using as ackbo_impl.elims(2)[OF subterm(2)] vars wei  by (smt Term.term.simps(5) is_VarE leD)
  }
  then have u_isFun: "is_Fun u" by blast
  then obtain f ss where u: "u = Fun f ss" by blast
  consider (i) "weight u = weight t" | (ii) "weight t < weight u" using subterm(2) by fastforce
  then show ?case
  proof cases
    case i
    then show ?thesis proof(cases "is_Var t")
      case True
      note t_isVar = True
      then have "ackbo_case0 u t" using vars i u_isFun
      proof(induct u)
        case (Var x)
        then show ?case by auto
      next
        case (Fun f ss)
        from True have wt: "weight t = w0" by fastforce
        {
          assume "length ss = 0"
          then have False using Fun(3) True by auto
        }
        then have "length ss \<ge> 1" by presburger 
        then have "sum_list (map weight ss) \<ge> w0"
          by (metis (no_types, lifting) distrib_left le_iff_add mult.right_neutral trans_le_add1 weight_lower_bound_list) 
        then have wf: "w(f,1) = 0" using Fun(4,5) wt apply auto
          by (metis One_nat_def \<open>1 \<le> length ss\<close> add_cancel_left_left leD le_add2 le_neq_implies_less mult.commute n_less_m_mult_n w0(2) weight_lower_bound_list) 
        then obtain su where s_u: "ss = [su]" using i u Fun(4)[symmetric] by (metis \<open>1 \<le> length ss\<close> impossible_Cons le_antisym less_one list.size(3) mult.commute mult.right_neutral nat_less_le nat_mult_le_cancel1 nat_neq_iff neq_Nil_conv w0(2) weight_lower_bound_n_arry_fun wt zero_le)
        then show ?case proof(cases "is_Var su")
          case True
          then have "su = t" using s_u Fun(3) t_isVar vars_term_ms.simps(1) by fastforce
          then show ?thesis using s_u ackbo_case0.simps[of "Fun f ss" t] t_isVar by blast
        next
          case False
          then obtain h hs where su: "su = Fun h hs" by blast
          {
            assume "length hs = 0"
            then have False using Fun(3) True s_u su by auto
          }
          then have "length hs \<ge> 1" by presburger 
          then have "sum_list (map weight hs) \<ge> w0"
            by (metis (no_types, lifting) distrib_left le_iff_add mult.right_neutral trans_le_add1 weight_lower_bound_list) 
          then have "w(h,1) = 0" using Fun(4,5) wt su s_u wf apply auto
            by (metis One_nat_def \<open>1 \<le> length hs\<close> add_cancel_left_left le_add2 le_antisym local.wf mult.right_neutral nat_mult_le_cancel1 w0(2) weight_lower_bound_list) 
          then have feq: "h = f" using wf unique_unary_function_weight0 by blast
          then have "ackbo_case0 (Fun f hs) t" using Fun(1)[of su] t_isVar Fun(3,4) s_u wf su by auto
          then show ?thesis using ackbo_case0.simps[of "Fun f ss" t] s_u su feq by auto
        qed
      qed
      from case_0[OF vars i this] show ?thesis .  
    next
      case False
      then obtain g ts where t: "t = Fun g ts" by blast
      then consider (a) "pr_strict (f, length ss) (g,length ts)" | (b) "(f, length ss) = (g, length ts)" using subterm(2) i
        by (smt Term.term.simps(6) ackbo_impl.elims(2) nat_less_le u)
      then show ?thesis
      proof cases
        case a
        then show ?thesis using vars i u t case_1 ackbo_case1.simps(1) by blast
      next
        case b
        then consider (2) "(f \<notin> AC \<or> length ss \<noteq> 2)  \<and> fst (lex_ext (\<lambda>x y. (ackbo_impl x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length ss) ss ts)" |
                      (3) "ackbo_case3 ackbo_impl u t"
          using subterm(2) u t by (smt Term.term.simps(6) ackbo_impl.elims(2) i nat_less_le pr_irr)
        then show ?thesis proof cases
          case 2
          have "\<forall>s \<in> set ss. \<forall>x. ackbo_impl s x \<longrightarrow> ackbo s x" using subterm(1) u by blast 
          moreover have "ackbo_case2 ackbo_impl u t" using u t b 2 ackbo_case2.simps(1) by auto
          ultimately have "ackbo_case2 ackbo u t" using ackbo_case2_local_mono u t b by blast
          then show ?thesis using ackbo.simps i vars by blast 
        next
          case 3
          then have "ackbo_case3 ackbo u t" using subterm(1) u t by (metis ackbo_case3_cong ackbo_eq_impl_left) 
          then show ?thesis by (simp add: case_3 i vars) 
        qed
      qed
    qed
  next
    case ii
    then show ?thesis using vars case_w by auto
  qed
qed

lemma ackbo_eq_impl:
  shows "ackbo s t = ackbo_impl s t"
  using ackbo_eq_impl_left ackbo_eq_impl_right by blast

end

end