(*
Author:  Alexander Lochmann <alexander.lochmann@uibk.ac.at> (2018)
License: LGPL (see file COPYING.LESSER)
*)
theory Ackbo_Impl
  imports
    Ackbo_more
    Term_Order_Impl
    TRS.Term_Impl
    Aux.Multiset_Code
    Aux.Map_Choice
    Multiset_Extension_Impl
begin

(*TODO: move*)
\<comment> \<open>show function symbol together with arity\<close>
fun showsl_funa :: "('f::showl \<times> nat) \<Rightarrow> showsl"
  where
    "showsl_funa (f, n) = showsl f \<circ> showsl_lit (STR ''/'') \<circ> showsl n"

abbreviation ackbo where "ackbo \<equiv> weight_prec_ac.ackbo_impl"


(* for representing concrete instances of ACKBO,
   entries:
    ('f \<times> nat) represents the function symbol with associated arity
    (nat \<times> nat \<times> bool) the first entry represents the precedence where 0 is the lowest 
                        the second entry represents the weight of the symbol
                        the third entry encodes the type of the symbol, if True then it is an AC symbol
    the nat after the list is the value of w0
*)
type_synonym 'f prec_weight_ac_repr = "(('f \<times> nat) \<times> (nat \<times> nat \<times> bool)) list \<times> nat"

fun showsl_ackbo_repr :: "('f :: showl) prec_weight_ac_repr \<Rightarrow> showsl"
where
  "showsl_ackbo_repr (prs, w0) =
    showsl_lit (STR ''ACKBO with the following precedence and weight function:\<newline>'') \<circ>
    foldr (\<lambda>(fn,(pr, w, ac)).
      showsl_lit (STR ''precedence('') \<circ> showsl_funa fn \<circ> showsl_lit (STR '') = '')
      \<circ> showsl pr \<circ> showsl_nl) prs \<circ>
    showsl_lit (STR ''precedence(_) = 0\<newline>\<newline>'') \<circ>
    foldr (\<lambda>(fn, (pr, w, ac)).
      showsl_lit (STR ''weight('') \<circ> showsl_funa fn \<circ> showsl_lit (STR '') = '')
      \<circ> showsl w \<circ> showsl_nl) prs \<circ>
    showsl_lit (STR ''weight(_) = '') \<circ> showsl (Suc w0) \<circ>
    showsl_lit (STR ''\<newline>w0 = '') \<circ> showsl w0 \<circ> showsl_nl \<circ> showsl_nl \<circ>
    showsl_list_gen (\<lambda>(fn, _). showsl_funa fn)
      (STR ''no AC function symbols'') \<comment> \<open>empty list\<close>
      (STR ''AC function symbols: '') \<comment> \<open>left of list\<close>
      (STR '', '') \<comment> \<open>separator for list items\<close>
      (STR '''') \<comment> \<open>right of list\<close>
      (filter (\<lambda>(_, (_, _, ac)). ac) prs) \<circ> showsl_nl \<circ>
    showsl_list_gen (\<lambda>(fn, _). showsl_funa fn)
      (STR ''no non AC function symbols'') \<comment> \<open>empty list\<close>
      (STR ''non AC function symbols: '') \<comment> \<open>left of list\<close>
      (STR '', '') \<comment> \<open>separator for list items\<close>
      (STR '''') \<comment> \<open>right of list\<close>
      (filter (\<lambda>(_, (_, _, ac)). \<not> ac) prs) \<circ> showsl_nl"

definition prec_ext :: "('a \<Rightarrow> (nat \<times> 'b) option) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool)"
  where "prec_ext prwm = (\<lambda> f g. case prwm f of
    Some pf \<Rightarrow> (case prwm g of Some pg \<Rightarrow> fst pf > fst pg | None \<Rightarrow> True)
    | None \<Rightarrow> False)"

lemma prec_ext_measure:
  fixes prwm :: "('a \<Rightarrow> (nat \<times> 'b) option)"
  defines "m \<equiv> fun_of_map_fun' prwm (\<lambda> _. 0) (Suc \<circ> fst)"
  shows  "prec_ext prwm fn gm = (m fn > m gm)"
proof
  assume s:"prec_ext prwm fn gm"
  from s[unfolded prec_ext_def] obtain pf where pf:"prwm fn = Some pf"
    by (metis (no_types, lifting) option.exhaust option.simps(4))
  show "m fn > m gm"
  proof(cases "prwm gm = None")
    case True
    with pf show "m fn > m gm" unfolding m_def by force
  next
    case False
    then obtain pg where pg:"prwm gm = Some pg" by auto
    from s[unfolded prec_ext_def pf pg option.cases] pf pg show "m fn > m gm" unfolding m_def by auto
  qed
next
  assume m:"m fn > m gm"
  from this[unfolded m_def] have pf:"prwm fn \<noteq> None"
    by (metis fun_of_map_fun'.simps not_less_zero option.simps(4))
  then obtain pf where pf:"prwm fn = Some pf" by auto
  show "prec_ext prwm fn gm" proof(cases "prwm gm = None")
    case False
    then obtain pg where pg:"prwm gm = Some pg" by auto
    from m[unfolded m_def] show ?thesis unfolding prec_ext_def pf pg option.cases using pf pg by force
  next
    case True
    show ?thesis unfolding prec_ext_def pf option.cases True by auto
  qed
qed

(* check whether partially defined prec and weight are admissible, convert to total weight and prec *)
definition
  prec_weight_ac_repr_to_prec_weight_funs :: "('f :: compare_order) prec_weight_ac_repr \<Rightarrow> ('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool) \<times> ('f \<times> nat \<Rightarrow> nat) \<times> nat \<times> 'f set" 
where
  "prec_weight_ac_repr_to_prec_weight_funs prw_w0 \<equiv>
    let 
      (prw, w0) = prw_w0;
      prwm    = ceta_map_of prw;
      w_fun   = fun_of_map_fun' prwm (\<lambda> _. Suc w0) (fst o snd);
      p_fun   = prec_ext prwm;
      acset = set (map (fst \<circ> fst) (filter (\<lambda> ((f,n),(p,w,ac)). ac) prw))
      in (p_fun, w_fun, w0, acset)"

definition
  prec_weight_ac_repr_to_prec_weight :: "('f :: {showl,compare_order}) prec_weight_ac_repr \<Rightarrow> showsl check \<times> ('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool) \<times> ('f \<times> nat \<Rightarrow> nat) \<times> nat \<times> 'f set" 
where
  "prec_weight_ac_repr_to_prec_weight prw_w0 \<equiv>
    let
      (p_fun, w_fun, w0, acset) = prec_weight_ac_repr_to_prec_weight_funs prw_w0;
      (prw, w0) = prw_w0;
      fs      = map fst prw;
      cw_okay = check_allm (\<lambda> fn. check (snd fn = 0 \<longrightarrow> w_fun fn \<ge> w0) 
          (showsl_lit (STR ''weight of constant '') o showsl (fst fn) o showsl_lit (STR '' must be at least w0''))) fs;
      adm     = check_allm (\<lambda> fn. check (snd fn = 1 \<longrightarrow> w_fun fn = 0 \<longrightarrow> (list_all (\<lambda> x. p_fun fn x \<or> x = fn) fs)) 
         (showsl_lit (STR ''unary symbol '') o showsl (fst fn) o showsl_lit (STR '' with weight 0 does not have maximal precedence''))) (map fst prw);
      irr     = check_allm (\<lambda> fn. check (\<not> p_fun fn fn) (showsl_lit (STR ''function symbol '') o showsl (fst fn) o showsl_lit (STR '' violates irreflexibity''))) fs;
      ok      = (do {
        check (w0 > 0) (showsl_lit (STR ''w0 must be larger than 0''));
        adm;
        cw_okay;
        irr
      })
      in (ok, p_fun, w_fun, w0, acset)"

lemma prec_weight_ac_repr_to_prec_weight: 
  assumes ok: "prec_weight_ac_repr_to_prec_weight prw_w0 = (succeed,p, w, w0, acset)"
  shows "admissible_weight_fun_ac w w0 p"
proof -
  have id: "\<And> b. (b = succeed) = isOK(b)" by auto
  note defs = prec_weight_ac_repr_to_prec_weight_def prec_weight_ac_repr_to_prec_weight_funs_def
  obtain prw w0' where prw_w0: "prw_w0 = (prw,w0')" by force
  obtain cs where cs: "cs = filter (\<lambda> fn. snd fn = 0 \<and> w fn = w0') (map fst prw)" by auto
  note ok' = ok[unfolded defs prw_w0 Let_def split]
  from ok' have w0': "w0' = w0" by simp
  note ok = ok'[unfolded w0']
  obtain fs where fs: "fs = set (map fst prw)" by auto
  note ok = ok[simplified, unfolded id, simplified]
  have w: "w = (\<lambda> fn. case map_of prw fn of Some pw \<Rightarrow> (fst o snd) pw | None \<Rightarrow> Suc w0)" 
    by (rule ext, insert ok, auto)
  from ok have p:"p = prec_ext (map_of prw)" by auto
  from ok have cw_okay: "\<And> fn. fn \<in> fs \<Longrightarrow> snd fn = 0 \<Longrightarrow> w fn \<ge> w0" unfolding w fs by auto
  from ok have "\<And> fn. fn \<in> fs \<Longrightarrow> snd fn = 1 \<longrightarrow> w fn = 0 \<longrightarrow> list_all (\<lambda> x. p fn x \<or> x = fn) (map fst prw)"
    unfolding fs w set_map comp_apply by auto
  hence adm: "\<And> fn gm. fn \<in> fs \<Longrightarrow> gm \<in> fs \<Longrightarrow> snd fn = 1 \<Longrightarrow> w fn = 0 \<Longrightarrow> p fn gm \<or> fn = gm"
    using fs unfolding list.pred_set by force
  from ok have irr:"\<And> fn. fn \<in> fs \<Longrightarrow> \<not> p fn fn" by (meson less_irrefl prec_ext_measure)
  from ok have w0: "w0 > 0" by simp
  {
    fix fn
    assume "fn \<notin> fs"
    hence fn: "\<And> e. (fn,e) \<notin> set prw" unfolding fs by force
    with map_of_SomeD[of prw fn] have l: "map_of prw fn = None" 
      by (cases "map_of prw fn", auto)
    hence "w fn = Suc w0" unfolding p w by auto
  } note not_fs = this
  show ?thesis
  proof
    show "w0 > 0" by (rule w0)
  next
    fix f
    show "w0 \<le> w (f,0)"
    proof (cases "(f,0) \<in> fs")
      case False
      from not_fs[OF this] show ?thesis by simp
    next
      case True
      from cw_okay[OF this] show ?thesis by simp
    qed
  next
    fix f and h ::"('a\<times>nat)"
    assume ass:"w (f,1) = 0 \<and> (f,1) \<noteq> h"
    then have wf: "w (f,1) = 0" by auto
    from ass have nt_eq: "(f,1) \<noteq> h" by auto
    obtain g n where h: "h = (g,n)" using ce_trs.cases by blast 
    with not_fs[of "(f,1)"] have f_fs:"(f,1) \<in> fs" using wf by auto
    obtain pf where f_some:"map_of prw (f, 1) = Some pf"
      using weak_map_of_SomeI[of "(f,1)" _ prw] f_fs[unfolded fs] by fastforce
    show "p (f,1) h"
    proof(cases "(g,n) \<in> fs")
      case False
      then have g_none:"map_of prw (g, n) = None" unfolding fs map_of_eq_None_iff by force
      from not_fs[OF False] show ?thesis unfolding p prec_ext_def g_none f_some h by fastforce
    next
      case True
      from True have "p (g,n) \<in> set (map p (map fst prw))" unfolding fs by force
      from adm[OF f_fs True _ wf] show ?thesis using h nt_eq by auto 
    qed
  next
    let ?m = "fun_of_map_fun' (map_of prw) (\<lambda> _. 0) (Suc \<circ> fst)"
    show "SN {(fn, gm). p fn gm}" unfolding p prec_ext_measure
      using SN_inv_image[OF SN_nat_gt, unfolded inv_image_def, of ?m] by fast
  next
    fix g
    show "\<not> p g g" using irr unfolding p prec_ext_def
      by (metis (no_types, lifting) fs map_of_eq_None_iff option.simps(4) set_map)
  next
    fix s t u
    assume "p s t \<and> p t u"
    then show "p s u" unfolding p prec_ext_def by (smt case_optionE nat_int_comparison(2) option.simps(4) option.simps(5))
  qed
qed


(* checking ackbo constraints: invoke ackbo, and throw error message if terms are not in relation *)
definition
  ackbo_strict :: "('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f set) \<Rightarrow> ('f :: showl,'v :: showl)rule \<Rightarrow> showsl check"
where
  "ackbo_strict pr w w0 acset \<equiv> \<lambda>(s, t). check (ackbo w w0 pr acset s t) (showsl_lit (STR ''could not orient '')  o showsl s o showsl_lit (STR '' >ACKBO '') o showsl t o showsl_nl)"

lemma ackbo_strict[simp]: "isOK (ackbo_strict pr w w0 ac st) = ackbo w w0 pr ac (fst st) (snd st)" unfolding ackbo_strict_def by (cases st, auto)

definition
  ackbo_nstrict :: "('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f set) \<Rightarrow> ('f :: showl,'v :: showl)rule \<Rightarrow> showsl check"
where
  "ackbo_nstrict pr w w0 acset \<equiv> \<lambda>(s, t). check (ackbo w w0 pr acset s t \<or> (s,t) \<in> (acstep acset acset)\<^sup>\<leftrightarrow>\<^sup>*) 
    (showsl_lit (STR ''could not orient '')  o showsl s o showsl_lit (STR '' >=ACKBO '') o showsl t o showsl_nl)"

lemma ackbo_nstrict[simp]: "isOK (ackbo_nstrict pr w w0 ac st) = (ackbo w w0 pr ac (fst st) (snd st) \<or> (fst st, snd st) \<in> (acstep ac ac)\<^sup>\<leftrightarrow>\<^sup>*)" unfolding ackbo_nstrict_def by (cases st, auto)

fun create_ACKBO_redtriple :: "(('f :: showl) prec_weight_ac_repr \<Rightarrow> 'g prec_weight_ac_repr) \<Rightarrow> 'f prec_weight_ac_repr \<Rightarrow> ('g :: {showl,compare_order},'v :: showl)redtriple"
where "create_ACKBO_redtriple f_to_g pr = (let (ch,p,w,w0,ac) = prec_weight_ac_repr_to_prec_weight (f_to_g pr);
   ns = ackbo_nstrict p w w0 ac;
   s =  ackbo_strict p w w0 ac
    in
  \<lparr> redtriple.valid = ch, 
    s = s,
    ns = ns,
    nst = ns,
    af = full_af,
    mono_af = full_af,
    mono = (\<lambda> _. succeed),
    desc = showsl_ackbo_repr pr,
    not_ws_ns = Some [],
    cpx = no_complexity_check\<rparr>)"

(* register ACKBO as executable reduction triple *)
interpretation ACKBO_redpair: generic_redtriple_impl "create_ACKBO_redtriple (f_to_g :: ('f :: showl)prec_weight_ac_repr \<Rightarrow> ('g :: {showl,compare_order})prec_weight_ac_repr)"
proof
  fix "prw" :: "'f prec_weight_ac_repr"
   and s_list ns_list nst_list :: "('g,'v :: showl)rule list"
  let ?rp = "create_ACKBO_redtriple f_to_g prw :: ('g,'v)redtriple"
  let ?af = "redtriple.af ?rp :: ('g af)"
  let ?af' = "redtriple.mono_af ?rp :: ('g af)"
  let ?pr = "prec_weight_ac_repr_to_prec_weight (f_to_g prw)"
  let ?cpx = "redtriple.cpx ?rp"
  let ?cpx' = "\<lambda> cm cc. isOK(?cpx cm cc)"
  let ?all = "s_list @ ns_list @ nst_list"
  let ?ws = "redtriple.not_ws_ns ?rp"
  obtain ch "pr" w w0 ac where id: "?pr = (ch,pr,w,w0,ac)" by (cases ?pr, force)
  assume valid: "isOK(redtriple.valid ?rp)"
   and stri: "isOK(check_allm (redtriple.s ?rp) s_list)"
   and nstri: "isOK (check_allm (redtriple.ns ?rp) ns_list)"
   and nstri_top: "isOK (check_allm (redtriple.nst ?rp) nst_list)"
  have af: "?af = full_af" "?af' = full_af" by (simp_all add: id Let_def)
  from valid have "isOK(ch)" by (simp add: id Let_def)
  hence ch: "ch = succeed" by (cases ch, auto)
  have cpx: "?cpx' = no_complexity" by (simp add: id Let_def)
  note id = id[unfolded ch cpx]
  interpret admissible_weight_fun_ac w w0 pr ac
    by (rule prec_weight_ac_repr_to_prec_weight[OF id])
  from ackbo_redpair
  interpret mono_ce_af_redtriple_order ackbo_S NS NS full_af .
  show "\<exists> S NS NST. cpx_ce_af_redtriple_order S NS NST ?af ?af' ?cpx' 
    \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST 
    \<and> not_ws_info NS (redtriple.not_ws_ns ?rp)
    \<and> (isOK (redtriple.mono ?rp ?all) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?af \<and> ctxt.closed S)"
    unfolding af cpx
  proof (intro exI conjI impI)
    show "cpx_ce_af_redtriple_order ackbo_S NS NS full_af full_af no_complexity"
      by (unfold_locales, rule ctxt_closed_imp_af_monotone[OF ackbo_S_mono])
    show "mono_ce_af_redtriple_order ackbo_S NS NS full_af" ..
    show "ctxt.closed ackbo_S" by (rule ackbo_S_mono)
    show "set s_list \<subseteq> ackbo_S" using stri apply auto
      by (metis (no_types, lifting) ackbo_eq_impl_right ackbo_strict id old.prod.case prod.sel(1) prod.sel(2) redtriple.select_convs(2)) 
    show "set ns_list \<subseteq> NS" using nstri apply auto
      by (metis (no_types, lifting) ackbo_eq_impl_right ackbo_nstrict case_prod_conv id prod.sel(1) prod.sel(2) redtriple.select_convs(3)) 
    show "set nst_list \<subseteq> NS" using nstri_top apply auto
      by (metis (no_types, lifting) ackbo_eq_impl_right ackbo_nstrict id old.prod.case prod.sel(1) prod.sel(2) redtriple.select_convs(4))
    show "not_ws_info NS ?ws"  by (simp add: id Let_def, intro allI ws_fun_argI NS_supteq, auto)
  qed
qed

context weight_prec_ac
begin
lemma ackbo_impl:
 "ackbo_impl s t = 
     (if (vars_term_ms t \<subseteq># vars_term_ms s \<and> weight t \<le> weight s)
      then if (weight t < weight s)
           then True
           else (case s of
                 Var y \<Rightarrow> False 
                 |Fun f ss \<Rightarrow> (case t of Var x \<Rightarrow> True
                               |Fun g ts \<Rightarrow> 
                                  if pr_strict (f,length ss) (g,length ts)
                                  then True
                                  else if (f, length ss) = (g, length ts)
                                  then if (f \<notin> AC \<or> length ss \<noteq> 2)
                                       then fst (lex_ext (\<lambda>x y. (ackbo_impl x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)) (length ss) ss ts)
                                       else let (S,T) = (actop f (Fun f ss), actop f (Fun g ts)) in
                                              ackbo_case3_filtered S T (f,length ss) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) ackbo_impl
                                else False))
       else False)"
proof (cases "is_Var s")
  case True
  then show ?thesis by auto
next
  case False
  note s_fun = False
  then show ?thesis
  proof (cases "is_Var t")
    case True
    then show ?thesis using False apply (auto) apply (fastforce)
      by fastforce
  next
    case False
    note t_fun = False
    then obtain f ss where s:"s = Fun f ss" using s_fun by auto
    then obtain g ts where t:"t = Fun g ts" using t_fun by auto
    then show ?thesis
    proof (cases "pr_strict (f, length ss) (g, length ts)")
      case True
      then show ?thesis using s t by auto
    next
      case False
      note nt_pr = False
      then show ?thesis
      proof (cases "(f, length ss) \<noteq> (g, length ts)")
        case True
        then show ?thesis using s t nt_pr
          by (smt Term.term.simps(6) ackbo_impl.simps)
      next
        case False
        then have "(f, length ss) = (g, length ts)" by auto
        note eq_root = this
        then show ?thesis
        proof (cases "f \<notin> AC \<or> length ss \<noteq> 2")
          case True
          then show ?thesis using s t eq_root nt_pr by (smt Term.term.simps(6) ackbo_impl.simps)
        next
          case False
          then have "f \<in> AC \<and> length ss = 2" by auto
          then show ?thesis unfolding mulextp_def smulextp_def nsmulextp_def ackbo_impl.simps[of s t] ackbo_case3.simps(1)[of ackbo_impl f ss g ts]
            using s t eq_root nt_pr apply (simp del: ackbo_impl.simps)
            by meson
        qed
      qed
    qed
  qed
qed

end

declare weight_fun.weight.simps[code]
declare weight_prec_ac.ackbo_impl[code]

end
